/*-
 * Copyright (c) 2013 Tomasz Walaszek
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <fcntl.h>
#include <kvm.h>

#include <limits.h>
#include <paths.h>

#include <err.h>
#include <errno.h>

#include <sys/param.h>
#include <sys/sysctl.h>
#include <sys/user.h>

#include <sys/queue.h>

extern int errno;

int
compare(const void *a, const void *b)
{
	const struct kinfo_proc *ka = a;
	const struct kinfo_proc *kb = b;

	pid_t _a = ka->ki_pid;
	pid_t _b = kb->ki_pid;

	if (_a < _b)
		return -1;
	else if (_a == _b)
		return 0;
	else
		return 1;
}

int
count_jail_procs(struct kinfo_proc *kp, int jail, int cnt)
{
	int count = 0;

	for (int i = 0; i < cnt; i++) {
		if (kp->ki_jid == jail || kp->ki_pid == 1) {
			count++;
		}
		kp++;
	}
	return count;
}

struct kinfo_proc *
get_jail_procs(struct kinfo_proc *kp, int jail, int cnt)
{
	struct kinfo_proc *local_kp, *ret;

	if ((local_kp = malloc(sizeof(struct kinfo_proc) * cnt)) == NULL)
		errx(1, "%s", strerror(errno));

	ret = local_kp;
	for (int i = 0; i < cnt; i++) {
		if (kp->ki_jid == jail || kp->ki_pid == 1) {
			memcpy(local_kp, kp, sizeof(struct kinfo_proc));
			local_kp++;
		}
		kp++;
	}

	return ret;
}

void
args(kvm_t * kd, struct kinfo_proc *kp)
{
	int  i;
	char **arg;

	arg = kvm_getargv(kd, kp, 0);
	if (arg == NULL) {
		printf(" %s", kp->ki_comm);		
		return;
	}

	i = 0;
	while (arg[i] != NULL) {
		printf(" %s", arg[i]);
		i++;
	}
}


int
child(struct kinfo_proc *kp, int pid, int cnt, int *seen, int *how, int br, int lg, kvm_t * kd)
{
	struct kinfo_proc *local_kp, *for_free;
	char              *bre;

	if ((local_kp = malloc(sizeof(struct kinfo_proc) * cnt)) == NULL)
		errx(1, "%s", strerror(errno));

	for_free = local_kp;

	if ((bre = malloc(sizeof(char) * br)) == NULL)
		errx(1, "%s", strerror(errno));

	for (int i = 0; i < br; i++) {
		strcat(bre, " ");
	}

	local_kp = memcpy(local_kp, kp, sizeof(struct kinfo_proc) * cnt);

	for (int i = 0; i < cnt; i++) {
foo:
		if (local_kp->ki_ppid == pid) {
			for (int j = 0; j <= cnt; j++) {
				if (local_kp->ki_pid == seen[j]) {
					i++;
					local_kp++;
					goto foo;
				}
			}
			printf("%s%d", bre, local_kp->ki_pid);
			if (lg)
				args(kd, local_kp);
			else
				printf(" %s", local_kp->ki_comm);
			printf("\n");
			seen[(*how)++] = local_kp->ki_pid;
			child(local_kp, local_kp->ki_pid, cnt, seen, how, br + 2, lg, kd);
		}
		local_kp++;
	}
	free(for_free);
	return 0;
}

void
usage()
{
	printf("usage: ptree\n");
	printf("-j <jail number>\n");
	printf("-p <pid number>\n");
	printf("-l - show processes with thier arguments\n");
	printf("no argument will show all processes\n");
	exit(0);
}

int
main(int argc, char *argv[])
{
	char	       c, errbuf[_POSIX2_LINE_MAX];
	const char     *nlistf, *memf;
	char           *jailf, *cpid, *end;

	int	       cnt, how, br, i, jail, count, lg;
	int            *seen;

	pid_t		pid;

	kvm_t          *kd;

	struct kinfo_proc *kp;

	how = br = i = jail = lg = 0;
	jailf = cpid = NULL;

	while ((c = getopt(argc, argv, "lhj:p:")) != -1) {
		switch (c) {
		case 'l':
			lg = 1;
			break;
		case 'j':
			jailf = optarg;
			break;
		case 'p':
			cpid = optarg;
			break;
		case 'h':
			usage();
		case '?':
			usage();
		}
	}


	nlistf = NULL;
	memf = _PATH_DEVNULL;

	kd = kvm_openfiles(nlistf, memf, NULL, O_RDONLY, errbuf);
	if (kd == 0)
		errx(1, "%s", errbuf);

	kp = kvm_getprocs(kd, KERN_PROC_PROC, 0, &cnt);
	if (kp == NULL)
		errx(1, "Cant get processes");

	if (jailf) {
		jail = (int)strtol(jailf, &end, 10);
		if (end == jailf || *end != '\0' || errno == ERANGE)
			errx(1, "Invalid jail id %s", jailf);
		count = count_jail_procs(kp, jail, cnt);
		if (count == 0 || jail == 0 || count == 1)
			errx(2, "Jail %d does not exist", jail);

		kp = get_jail_procs(kp, jail, cnt);
		cnt = count;
	}
	if (cpid) {
		pid = (int)strtol(cpid, &end, 10);
		if (end == cpid || *end != '\0' || errno == ERANGE)
			errx(1, "Invalid process id %s", cpid);
	}
	if ((seen = malloc(sizeof(int) * cnt)) == NULL)
		errx(1, "%s", strerror(errno));

	qsort(kp, cnt, sizeof(struct kinfo_proc), compare);

	for (i = 0; i < cnt; i++) {
		if (cpid) {
			if (pid != kp->ki_pid) {
				kp++;
				continue;
			}
		}
		for (int j = 0; j < how; j++) {
			if (kp->ki_pid == seen[j]) {
				kp++;
				br++;
				break;
			}
		}
		if (br) {
			br = 0;
			continue;
		}
		printf("%d", kp->ki_pid);
		if (lg) 
			args(kd, kp);
		else 
			printf(" %s", kp->ki_comm);			

		printf("\n");

		child(kp, kp->ki_pid, cnt, seen, &how, 2, lg, kd);
		if (cpid) {
			break;
		}
		kp++;
	}

	return 0;
}
